@@ -7,12 +7,13 @@ sticking incoming connections onto a Queue::
 
     server = HTTPServer(...)
     server.start()
-    ->  while True:
-            tick()
-            # This blocks until a request comes in:
-            child = socket.accept()
-            conn = HTTPConnection(child, ...)
-            server.requests.put(conn)
+    ->  serve()
+        while ready:
+            _connections.run()
+                while not stop_requested:
+                    child = socket.accept()  # blocks until a request comes in
+                    conn = HTTPConnection(child, ...)
+                    server.process_conn(conn)  # adds conn to threadpool
 
 Worker threads are kept in a pool and poll the Queue, popping off and then
 handling each connection in turn. Each connection can consist of an arbitrary
@@ -181,7 +182,7 @@ class HeaderReader:
     Interface and default implementation.
     """
 
-    def __call__(self, rfile, hdict=None):
+    def __call__(self, rfile, hdict=None):  # noqa: C901  # FIXME
         """
         Read headers from the given stream into the given header dict.
 
@@ -726,7 +727,7 @@ class HTTPRequest:
 
         self.ready = True
 
-    def read_request_line(self):
+    def read_request_line(self):  # noqa: C901  # FIXME
         """Read and parse first line of the HTTP request.
 
         Returns:
@@ -957,7 +958,7 @@ class HTTPRequest:
 
         return True
 
-    def read_request_headers(self):
+    def read_request_headers(self):  # noqa: C901  # FIXME
         """Read ``self.rfile`` into ``self.inheaders``.
 
         Ref: :py:attr:`self.inheaders <HTTPRequest.outheaders>`.
@@ -1128,7 +1129,7 @@ class HTTPRequest:
         else:
             self.conn.wfile.write(chunk)
 
-    def send_headers(self):
+    def send_headers(self):  # noqa: C901  # FIXME
         """Assert, process, and send the HTTP response message-headers.
 
         You must set ``self.status``, and :py:attr:`self.outheaders
@@ -1162,7 +1163,7 @@ class HTTPRequest:
         # Override the decision to not close the connection if the connection
         # manager doesn't have space for it.
         if not self.close_connection:
-            can_keep = self.server.connections.can_add_keepalive_connection
+            can_keep = self.server.can_add_keepalive_connection
             self.close_connection = not can_keep
 
         if b'connection' not in hkeys:
@@ -1178,7 +1179,9 @@ class HTTPRequest:
         if (b'Connection', b'Keep-Alive') in self.outheaders:
             self.outheaders.append((
                 b'Keep-Alive',
-                u'timeout={}'.format(self.server.timeout).encode('ISO-8859-1'),
+                u'timeout={connection_timeout}'.
+                format(connection_timeout=self.server.timeout).
+                encode('ISO-8859-1'),
             ))
 
         if (not self.close_connection) and (not self.chunked_read):
@@ -1231,9 +1234,7 @@ class HTTPConnection:
     peercreds_resolve_enabled = False
 
     # Fields set by ConnectionManager.
-    closeable = False
     last_used = None
-    ready_with_data = False
 
     def __init__(self, server, sock, makefile=MakeFile):
         """Initialize HTTPConnection instance.
@@ -1262,7 +1263,7 @@ class HTTPConnection:
             lru_cache(maxsize=1)(self.get_peer_creds)
         )
 
-    def communicate(self):
+    def communicate(self):  # noqa: C901  # FIXME
         """Read each request and respond appropriately.
 
         Returns true if the connection should be kept open.
@@ -1354,6 +1355,9 @@ class HTTPConnection:
 
         if not self.linger:
             self._close_kernel_socket()
+            # close the socket file descriptor
+            # (will be closed in the OS if there is no
+            # other reference to the underlying socket)
             self.socket.close()
         else:
             # On the other hand, sometimes we want to hang around for a bit
@@ -1465,18 +1469,20 @@ class HTTPConnection:
         return group
 
     def _close_kernel_socket(self):
-        """Close kernel socket in outdated Python versions.
+        """Terminate the connection at the transport level."""
+        # Honor ``sock_shutdown`` for PyOpenSSL connections.
+        shutdown = getattr(
+            self.socket, 'sock_shutdown',
+            self.socket.shutdown,
+        )
 
-        On old Python versions,
-        Python's socket module does NOT call close on the kernel
-        socket when you call socket.close(). We do so manually here
-        because we want this server to send a FIN TCP segment
-        immediately. Note this must be called *before* calling
-        socket.close(), because the latter drops its reference to
-        the kernel socket.
-        """
-        if six.PY2 and hasattr(self.socket, '_sock'):
-            self.socket._sock.close()
+        try:
+            shutdown(socket.SHUT_RDWR)  # actually send a TCP FIN
+        except errors.acceptable_sock_shutdown_exceptions:
+            pass
+        except socket.error as e:
+            if e.errno not in errors.acceptable_sock_shutdown_error_codes:
+                raise
 
 
 class HTTPServer:
@@ -1518,6 +1524,11 @@ class HTTPServer:
     timeout = 10
     """The timeout in seconds for accepted connections (default 10)."""
 
+    expiration_interval = 0.5
+    """The interval, in seconds, at which the server checks for
+    expired connections (default 0.5).
+    """
+
     version = 'Cheroot/{version!s}'.format(version=__version__)
     """A version string for the HTTPServer."""
 
@@ -1587,7 +1598,6 @@ class HTTPServer:
         self.requests = threadpool.ThreadPool(
             self, min=minthreads or 1, max=maxthreads,
         )
-        self.connections = connections.ConnectionManager(self)
 
         if not server_name:
             server_name = self.version
@@ -1613,29 +1623,29 @@ class HTTPServer:
             'Threads Idle': lambda s: getattr(self.requests, 'idle', None),
             'Socket Errors': 0,
             'Requests': lambda s: (not s['Enabled']) and -1 or sum(
-                [w['Requests'](w) for w in s['Worker Threads'].values()], 0,
+                (w['Requests'](w) for w in s['Worker Threads'].values()), 0,
             ),
             'Bytes Read': lambda s: (not s['Enabled']) and -1 or sum(
-                [w['Bytes Read'](w) for w in s['Worker Threads'].values()], 0,
+                (w['Bytes Read'](w) for w in s['Worker Threads'].values()), 0,
             ),
             'Bytes Written': lambda s: (not s['Enabled']) and -1 or sum(
-                [w['Bytes Written'](w) for w in s['Worker Threads'].values()],
+                (w['Bytes Written'](w) for w in s['Worker Threads'].values()),
                 0,
             ),
             'Work Time': lambda s: (not s['Enabled']) and -1 or sum(
-                [w['Work Time'](w) for w in s['Worker Threads'].values()], 0,
+                (w['Work Time'](w) for w in s['Worker Threads'].values()), 0,
             ),
             'Read Throughput': lambda s: (not s['Enabled']) and -1 or sum(
-                [
+                (
                     w['Bytes Read'](w) / (w['Work Time'](w) or 1e-6)
                     for w in s['Worker Threads'].values()
-                ], 0,
+                ), 0,
             ),
             'Write Throughput': lambda s: (not s['Enabled']) and -1 or sum(
-                [
+                (
                     w['Bytes Written'](w) / (w['Work Time'](w) or 1e-6)
                     for w in s['Worker Threads'].values()
-                ], 0,
+                ), 0,
             ),
             'Worker Threads': {},
         }
@@ -1719,7 +1729,7 @@ class HTTPServer:
             self.stop()
             raise
 
-    def prepare(self):
+    def prepare(self):  # noqa: C901  # FIXME
         """Prepare server to serving requests.
 
         It binds a socket's port, setups the socket to ``listen()`` and does
@@ -1781,6 +1791,9 @@ class HTTPServer:
         self.socket.settimeout(1)
         self.socket.listen(self.request_queue_size)
 
+        # must not be accessed once stop() has been called
+        self._connections = connections.ConnectionManager(self)
+
         # Create worker threads
         self.requests.start()
 
@@ -1791,22 +1804,15 @@ class HTTPServer:
         """Serve requests, after invoking :func:`prepare()`."""
         while self.ready:
             try:
-                self.tick()
+                self._connections.run(self.expiration_interval)
             except (KeyboardInterrupt, SystemExit):
                 raise
             except Exception:
                 self.error_log(
-                    'Error in HTTPServer.tick', level=logging.ERROR,
+                    'Error in HTTPServer.serve', level=logging.ERROR,
                     traceback=True,
                 )
 
-            if self.interrupt:
-                while self.interrupt is True:
-                    # Wait for self.stop() to complete. See _set_interrupt.
-                    time.sleep(0.1)
-                if self.interrupt:
-                    raise self.interrupt
-
     def start(self):
         """Run the server forever.
 
@@ -1831,6 +1837,19 @@ class HTTPServer:
         finally:
             self.stop()
 
+    @property
+    def can_add_keepalive_connection(self):
+        """Flag whether it is allowed to add a new keep-alive connection."""
+        return self.ready and self._connections.can_add_keepalive_connection
+
+    def put_conn(self, conn):
+        """Put an idle connection back into the ConnectionManager."""
+        if self.ready:
+            self._connections.put(conn)
+        else:
+            # server is shutting down, just close it
+            conn.close()
+
     def error_log(self, msg='', level=20, traceback=False):
         """Write error message to log.
 
@@ -1858,7 +1877,7 @@ class HTTPServer:
         self.bind_addr = self.resolve_real_bind_addr(sock)
         return sock
 
-    def bind_unix_socket(self, bind_addr):
+    def bind_unix_socket(self, bind_addr):  # noqa: C901  # FIXME
         """Create (or recreate) a UNIX socket object."""
         if IS_WINDOWS:
             """
@@ -2021,20 +2040,13 @@ class HTTPServer:
 
         return bind_addr
 
-    def tick(self):
-        """Accept a new connection and put it on the Queue."""
-        if not self.ready:
-            return
-
-        conn = self.connections.get_conn(self.socket)
-        if conn:
-            try:
-                self.requests.put(conn)
-            except queue.Full:
-                # Just drop the conn. TODO: write 503 back?
-                conn.close()
-
-        self.connections.expire()
+    def process_conn(self, conn):
+        """Process an incoming HTTPConnection."""
+        try:
+            self.requests.put(conn)
+        except queue.Full:
+            # Just drop the conn. TODO: write 503 back?
+            conn.close()
 
     @property
     def interrupt(self):
@@ -2047,14 +2059,21 @@ class HTTPServer:
         self._interrupt = True
         self.stop()
         self._interrupt = interrupt
+        if self._interrupt:
+            raise self.interrupt
 
-    def stop(self):
+    def stop(self):  # noqa: C901  # FIXME
         """Gracefully shutdown a server that is serving forever."""
+        if not self.ready:
+            return  # already stopped
+
         self.ready = False
         if self._start_time is not None:
             self._run_time += (time.time() - self._start_time)
         self._start_time = None
 
+        self._connections.stop()
+
         sock = getattr(self, 'socket', None)
         if sock:
             if not isinstance(
@@ -2096,7 +2115,7 @@ class HTTPServer:
                 sock.close()
             self.socket = None
 
-        self.connections.close()
+        self._connections.close()
         self.requests.stop(self.shutdown_timeout)
 
 
